from __future__ import annotations

import dataclasses
from dataclasses import dataclass

from typing import Any, Dict, List, TYPE_CHECKING

from . import log, tools, utils

if TYPE_CHECKING:
    from pathlib import Path

    from . import settings, templates

_logger = log.get_logger()


@dataclass
class Project:
    tools: Dict[tools.ToolUid, tools.ToolScript] = dataclasses.field(default_factory=dict)

    def __post_init__(self) -> None:
        for uid, script in self.tools.items():
            # dataclass can't handle nesting, so deserealization has to be done mannualy
            if isinstance(script, dict):
                self.tools[uid] = tools.ToolScript(**script)


def create(
    project_file: Path,
    design_kind: templates.DesignKind,
    sources: List[str],
    user_settings: settings.UserSettings,
) -> Project:
    project_tools: Dict[tools.ToolUid, tools.ToolScript] = {}

    for uid, tool_settings in user_settings.tools.items():
        try:
            project_tools[uid] = tools.generate_script(tool_settings, design_kind, sources)
        except (KeyError, ValueError, NotImplementedError):
            pass

    if len(list(project_tools.keys())) == 0:
        raise ValueError(f"Can't find any suitable tool for the provided design mode '{design_kind}'")

    project = Project(tools=project_tools)
    _logger.debug(f"Created project: {project}")

    return project


def dump(file: Path, project: Project, **kwargs: Any) -> None:
    """Dump project to file"""
    if utils.is_write_allowed(file, bool(kwargs.get("query_force_yes", False))):
        utils.dump_json(file, dataclasses.asdict(project))
    _logger.info(f"Project was successfuly saved to '{file}'")


def load(file: Path) -> Project:
    """Load project from file"""
    data = utils.load_json(file)
    project = Project(**data)
    _logger.debug(f"Loaded project: {project}")
    _logger.info(f"Project was successfuly loaded from '{file}'")
    return project



from __future__ import annotations

import os
import shutil
import subprocess
import sys
from pathlib import Path
from typing import Dict, TYPE_CHECKING

from . import log, project, settings

if TYPE_CHECKING:
    from . import tools

_logger = log.get_logger()


def _patch_path(bin_dir: Path, env: Dict[str, str]) -> None:
    """Patch PATH variable inplace"""
    env_path = f"{bin_dir}:{env.get('PATH', '')}"
    env.setdefault("PATH", env_path)


def _prepare_work_dir(tool_uid: tools.ToolUid) -> Path:
    """Prepare working directory"""
    work_dir = Path(f"./{tool_uid}")
    _logger.info(f"Clear working directory '{work_dir}'")
    shutil.rmtree(work_dir, ignore_errors=True)
    work_dir.mkdir()
    return work_dir


def _exec(cmd: str, cwd: Path, bin_dir: Path, env: Dict[str, str]) -> None:
    """Execute system process"""
    if not cmd:
        _logger.warning("Command is empty. Nothing to do.")
        return

    # Prepare environment
    proc_env = os.environ.copy()
    proc_env.update(env)
    _patch_path(bin_dir, proc_env)

    # Run process
    kwargs = {
        "args": cmd,
        "cwd": cwd,
        "env": os.environ.copy(),
        "stdout": sys.stdout,
        "stderr": subprocess.STDOUT,
        "shell": True,
    }
    proc = subprocess.run(**kwargs)  # type: ignore
    if proc.returncode != 0:
        raise RuntimeError(f"Command '{cmd}' returned {proc.returncode}. Check the output above for diagnostics.")


def run(project: project.Project, settings: settings.UserSettings, tool_uid: tools.ToolUid, waves: bool) -> None:
    # Check that provided tool exists
    if tool_uid not in project.tools.keys():
        raise ValueError(
            f"Tool '{tool_uid}' was not found in your project file. Available tools: {list(project.tools.keys())}."
            " Check your global settings and project file, then run again."
        )

    # Prepare tool attributes
    tool_settings = settings.tools[tool_uid]
    tool_script = project.tools[tool_uid]
    work_dir = _prepare_work_dir(tool_uid)

    # Run tool
    _logger.info("Run compilation ...")
    for cmd in tool_script.build:
        _logger.info(f"  {cmd}")
        _exec(cmd=cmd, cwd=work_dir, bin_dir=tool_settings.bin_dir, env=tool_settings.env)

    _logger.info("Run simulation ...")
    for cmd in tool_script.sim:
        _logger.info(f"  {cmd}")
        _exec(cmd=cmd, cwd=work_dir, bin_dir=tool_settings.bin_dir, env=tool_settings.env)

    if waves:
        _logger.info("Show waves ...")
        for cmd in tool_script.waves:
            _logger.info(f"  {cmd}")
            _exec(cmd=cmd, cwd=work_dir, bin_dir=tool_settings.bin_dir, env=tool_settings.env)



from __future__ import annotations

import dataclasses
from dataclasses import dataclass

from typing import Any, Dict, TYPE_CHECKING

from . import log, tools, utils

if TYPE_CHECKING:
    from pathlib import Path

_logger = log.get_logger()


@dataclass
class UserSettings:
    tools: Dict[tools.ToolUid, tools.ToolSettings] = dataclasses.field(default_factory=dict)

    def __post_init__(self) -> None:
        for uid, settings in self.tools.items():
            # dataclass can't handle nesting, so deserealization has to be done mannualy
            if isinstance(settings, dict):
                self.tools[uid] = tools.ToolSettings(**settings)


def setup(app_dir: Path, user_settings_file: Path, **kwargs: Any) -> None:
    """Create settings and application folder for a first time"""
    # Prepare application directory
    _logger.info(f"Create application home ...'{app_dir}'")
    app_dir.mkdir(parents=True, exist_ok=True)

    # Prepare settings file
    _logger.info("Create default settings ...")
    _logger.info("  Try to find all tools available ...")
    tool_pool = {}
    for t in tools.ToolKind:
        bin_dir = tools.find_tool_dir(t)
        _logger.info(f"  {t}: {bin_dir}")
        if bin_dir:
            tool_pool[str(t)] = tools.ToolSettings(kind=t, bin_dir=bin_dir)
    user_settings = UserSettings(tools=tool_pool)

    # Try to save settings to file
    _logger.info(f"Save settings to '{user_settings_file}' ...")
    if utils.is_write_allowed(user_settings_file, bool(kwargs.get("query_force_yes", False))):
        dump(user_settings_file, user_settings)


def dump(file: Path, settings: UserSettings) -> None:
    """Dump user settings to file"""
    utils.dump_json(file, dataclasses.asdict(settings))
    _logger.info(f"Settings were successfuly dumped to '{file}'")


def load(file: Path) -> UserSettings:
    """Load user settings from file"""
    data = utils.load_json(file)
    settings = UserSettings(**data)
    _logger.debug(f"Loaded user settings: {settings}")
    _logger.info(f"Settings were successfuly loaded from '{file}'")
    return settings



import logging
import os


class _CustomFormatter(logging.Formatter):
    """Formatter based on https://stackoverflow.com/a/56944256"""

    grey = "\x1b[38;20m"
    cyan = "\x1b[34;20m"
    green = "\x1b[32;20m"
    yellow = "\x1b[33;20m"
    red = "\x1b[31;20m"
    bold_red = "\x1b[31;1m"
    reset = "\x1b[0m"
    generic_format = "[%(name)s][%(levelname)s] %(message)s"
    debug_format = "[%(name)s][%(levelname)s][%(filename)s:%(lineno)d] %(message)s"

    formats = {
        logging.DEBUG: green + debug_format + reset,
        logging.INFO: cyan + generic_format + reset,
        logging.WARNING: yellow + generic_format + reset,
        logging.ERROR: red + generic_format + reset,
        logging.CRITICAL: bold_red + generic_format + reset,
    }

    def format(self, record: logging.LogRecord) -> str:
        log_fmt = self.formats.get(record.levelno)
        formatter = logging.Formatter(log_fmt)
        return formatter.format(record)


def is_debug_en() -> bool:
    """Is debug enabled"""
    return "DEBUG" in os.environ


def init_logger() -> None:
    """Initialize global logger"""
    # Create logger
    logger = logging.getLogger("playhdl")
    logger.setLevel(logging.DEBUG)

    # Create console handler with custom formatting
    ch = logging.StreamHandler()
    ch.setLevel(logging.DEBUG if is_debug_en() else logging.INFO)
    ch.setFormatter(_CustomFormatter())
    logger.addHandler(ch)


def get_logger() -> logging.Logger:
    """Get package logger"""
    return logging.getLogger("playhdl")
